//-----------------------------------------------------------------------------
// Copyright (c) 2007 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

#include "mesh_utils.h"

Mesh::Mesh() : pMesh(0), pDeclaration(0)
{
}

Mesh::~Mesh()
{
    destroy();
}

bool Mesh::create(const char *pszFilename, IDirect3DDevice9 *pDevice)
{
    std::string::size_type offset;
    std::string directoryPath = pszFilename;
    
    if ((offset = directoryPath.find_last_of('\\')) != std::string::npos)
        directoryPath = directoryPath.substr(0, offset + 1);
    else if ((offset = directoryPath.find_last_of('/')) != std::string::npos)
        directoryPath = directoryPath.substr(0, offset + 1);
    else
        directoryPath = "";

    LPD3DXBUFFER pBuffer = 0;
    DWORD numMaterials = 0;
    HRESULT hr = 0;

    hr = D3DXLoadMeshFromX(pszFilename, D3DXMESH_MANAGED, pDevice,
            0, &pBuffer, 0, &numMaterials, &pMesh);

    if (FAILED(hr))
        return false;

    D3DXMATERIAL *pD3DXMaterials = reinterpret_cast<D3DXMATERIAL *>(
        pBuffer->GetBufferPointer());

    materials.resize(numMaterials);
    textures.resize(numMaterials);

    for (DWORD i = 0; i < numMaterials; ++i)
    {
        materials[i] = pD3DXMaterials[i].MatD3D;

        if (pD3DXMaterials[i].pTextureFilename)
        {
            hr = D3DXCreateTextureFromFile(pDevice,
                    pD3DXMaterials[i].pTextureFilename, &textures[i]);

            if (FAILED(hr))
            {
                std::string path = directoryPath + pD3DXMaterials[i].pTextureFilename;

                hr = D3DXCreateTextureFromFile(pDevice, path.c_str(), &textures[i]);

                if (FAILED(hr))
                {
                    pBuffer->Release();
                    destroy();
                    return false;
                }
            }
        }
        else
        {
            textures[i] = 0;
        }
    }

    pBuffer->Release();

    elements.resize(MAXD3DDECLLENGTH + 1);

    if (FAILED(pMesh->GetDeclaration(&elements[0])))
        return false;
    
    if (FAILED(pDevice->CreateVertexDeclaration(&elements[0], &pDeclaration)))
        return false;

    return true;
}

void Mesh::destroy()
{
    for (int i = 0; i < static_cast<int>(textures.size()); ++i)
    {
        if (textures[i])
            textures[i]->Release();
    }

    textures.clear();
    materials.clear();

    if (pDeclaration)
    {
        pDeclaration->Release();
        pDeclaration = 0;
    }

    if (pMesh)
    {
        pMesh->Release();
        pMesh = 0;
    }
}

int Mesh::getElementOffsetByUsage(D3DDECLUSAGE usage) const
{
    std::vector<D3DVERTEXELEMENT9>::const_iterator iter = elements.begin();
    std::vector<D3DVERTEXELEMENT9>::const_iterator end = elements.end();
    int offset = 0;
    bool found = false;

    while (iter != end)
    {
        offset += iter->Offset;

        if (iter->Usage == usage)
        {
            found = true;
            break;
        }
        
        ++iter;
    }

    return found ? offset : -1;
}

// The methods bounds(), normalize(), and scale() are based on source code
// from http://www.mvps.org/directx/articles/scalemesh9.htm

bool Mesh::bounds(D3DXVECTOR3 *pCenter, float *pRadius)
{
    if (!pMesh)
        return false;
    
    BYTE *pBytes = 0;
    DWORD dwNumVertices = pMesh->GetNumVertices();
    DWORD dwVertexSize = D3DXGetFVFVertexSize(pMesh->GetFVF());

    if (FAILED(pMesh->LockVertexBuffer(0, reinterpret_cast<LPVOID*>(&pBytes))))
        return false;

    HRESULT hr = D3DXComputeBoundingSphere(
                    reinterpret_cast<D3DXVECTOR3*>(pBytes),
                    dwNumVertices, dwVertexSize, pCenter, pRadius);

    if (FAILED(hr))
    {
        pMesh->UnlockVertexBuffer();
        return false;
    }

    if (FAILED(pMesh->UnlockVertexBuffer()))
        return false;
    
    return true;
}

bool Mesh::normalize(float scaleTo, bool center)
{
    float radius;
    D3DXVECTOR3 centerPos;

    if (FAILED(bounds(&centerPos, &radius)))
        return false;

    float scalingFactor = scaleTo / radius;
    D3DXVECTOR3 offset = (center) ? -centerPos : D3DXVECTOR3(0.0f, 0.0f, 0.0f);

    return scale(scalingFactor, &offset);
}

bool Mesh::rotate(float degrees, float x, float y, float z)
{
    DWORD dwNumVertices = pMesh->GetNumVertices();
    DWORD dwFVF = pMesh->GetFVF();
    DWORD dwVertexSize = D3DXGetFVFVertexSize(pMesh->GetFVF());
    BYTE *pData = 0;
        
    if (FAILED(pMesh->LockVertexBuffer(0, reinterpret_cast<LPVOID*>(&pData))))
        return false;
    
    BYTE *pByte = pData;
    D3DXVECTOR3 *pVector = 0;
    D3DXVECTOR3 axis(x, y, z);
    D3DXMATRIX mtx;
    
    D3DXMatrixRotationAxis(&mtx, &axis, D3DXToRadian(degrees));

    for (DWORD i = 0; i < dwNumVertices; ++i)
    {
        pVector = reinterpret_cast<D3DXVECTOR3*>(pByte);
        D3DXVec3TransformCoord(pVector, pVector, &mtx);
        pByte += dwVertexSize;
    }

    int normalOffset = getElementOffsetByUsage(D3DDECLUSAGE_NORMAL);

    if (normalOffset != -1)
    {
        pByte = pData + normalOffset;  

        for (DWORD i = 0; i < dwNumVertices; ++i)
        {
            pVector = reinterpret_cast<D3DXVECTOR3*>(pByte);
            D3DXVec3TransformNormal(pVector, pVector, &mtx);
            pByte += dwVertexSize;
        }
    }

    if (FAILED(pMesh->UnlockVertexBuffer()))
        return false;

    return true;
}

bool Mesh::scale(float scale, D3DXVECTOR3 *pOffset)
{
    D3DXVECTOR3 offset = (pOffset) ? *pOffset : D3DXVECTOR3(0.0f, 0.0f, 0.0f);
    DWORD dwNumVertices = pMesh->GetNumVertices();
    DWORD dwFVF = pMesh->GetFVF();
    DWORD dwVertexSize = D3DXGetFVFVertexSize(pMesh->GetFVF());
    BYTE *pBytes = 0;
    
    if (FAILED(pMesh->LockVertexBuffer(0, reinterpret_cast<LPVOID*>(&pBytes))))
        return false;
    
    D3DXVECTOR3 *pVector = 0;

    for (DWORD i = 0; i < dwNumVertices; ++i)
    {
        pVector = reinterpret_cast<D3DXVECTOR3*>(pBytes);

        *pVector += offset;
        
        pVector->x *= scale;
        pVector->y *= scale;
        pVector->z *= scale;

        pBytes += dwVertexSize;
    }

    if (FAILED(pMesh->UnlockVertexBuffer()))
        return false;

    return true;
}
